from __future__ import annotations

import logging
import struct
from typing import Optional, TYPE_CHECKING, Any

from dragodis.exceptions import NotExistError
from dragodis.ida.instruction import IDAInstruction
from dragodis.interface import Line, LineType, CommentType

if TYPE_CHECKING:
    from dragodis.ida.flat import IDAFlatAPI


logger = logging.getLogger(__name__)


# TODO: Fixup the exceptions used here. Switch to custom exceptions?


class IDALine(Line):

    def __init__(self, ida: IDAFlatAPI, addr: int):
        super().__init__(ida)
        self._ida = ida
        if addr < 0:
            raise NotExistError(f"Got negative address: {addr}")
        # IDA has no concept of a "line", so we'll keep track of the start address,
        # which IDA refers to as the "head".
        start_addr = self._ida._ida_bytes.get_item_head(addr)
        if start_addr == self._ida._BADADDR:
            raise NotExistError(f"Line at {hex(addr)} does not exist.")
        self._addr = start_addr
        self._name = None

    # TODO: Move these into utility module inside IDA
    def _iter_extra_comments(self, addr: int, start: int):
        end = self._ida._ida_lines.get_first_free_extra_cmtidx(addr, start)
        for idx in range(start, end):
            comment_line = self._ida._ida_lines.get_extra_cmt(addr, idx)
            yield comment_line or ""

    def _set_extra_comments(self, addr: int, start: int, comment: str):
        if not comment:
            self._ida._ida_lines.del_extra_cmt(addr, start)
        else:
            index = 0
            for index, comment_line in enumerate(comment.splitlines()):
                self._ida._ida_lines.update_extra_cmt(addr, start + index, comment_line)
            self._ida._ida_lines.del_extra_cmt(addr, start + (index + 1))

    @property
    def address(self):
        return self._addr

    @property
    def data(self) -> bytes:
        if not self.is_loaded:
            return b""
        # Call flat api to take advantage of memory caching.
        return self._ida.get_bytes(self.address, self.size)

    @data.setter
    def data(self, new_data: bytes):
        if not self.is_loaded:
            return
        type_ = self.type
        self.undefine()
        self._ida._ida_bytes.patch_bytes(self.address, new_data)
        self.type = type_

    def get_comment(self, comment_type=CommentType.eol) -> Optional[str]:
        if comment_type == CommentType.eol:
            return self._ida._ida_bytes.get_cmt(self.address, 0) or None
        elif comment_type == CommentType.repeatable:
            return self._ida._ida_bytes.get_cmt(self.address, 1) or None
        elif comment_type in (CommentType.anterior, CommentType.plate):
            return "\n".join(self._iter_extra_comments(self.address, self._ida._ida_lines.E_PREV)) or None
        elif comment_type == CommentType.posterior:
            return "\n".join(self._iter_extra_comments(self.address, self._ida._ida_lines.E_NEXT)) or None
        else:
            raise ValueError(f"Invalid comment type: {repr(comment_type)}")

    @property
    def name(self) -> Optional[str]:
        if not self._name:
            self._name = self._ida._ida_name.get_name(self.address) or None
        return self._name

    @name.setter
    def name(self, new_name: Optional[str]):
        flags = self._ida._ida_name.SN_NOCHECK
        if new_name:
            # only force if not clearing.
            flags |= self._ida._ida_name.SN_FORCE
        else:
            new_name = ""

        success = self._ida._ida_name.set_name(self.address, new_name, flags)
        if not success:
            raise ValueError(f"Failed to set {hex(self.address)} to {new_name}")
        self._name = None  # clear cache

        # If we reset the name for a string we need to undefine it
        # so IDA gives us the appropriate auto-generated name.
        # Otherwise, a data item original named like "aHello" will be reset to "asc_40C130"
        if not new_name and self.is_string:
            type_ = self.type
            self.undefine()
            self.type = type_

    @property
    def next(self) -> Optional["IDALine"]:
        addr = self._ida._idc.next_head(self.address)
        if addr != self._ida._BADADDR:
            return IDALine(self._ida, addr)

    @property
    def prev(self) -> Optional["IDALine"]:
        addr = self._ida._idc.prev_head(self.address)
        if addr != self._ida._BADADDR:
            return IDALine(self._ida, addr)

    def set_comment(self, comment: Optional[str], comment_type=CommentType.eol):
        # IDA takes an empty string to clear comments
        if not comment:
            comment = ""

        if comment_type == CommentType.eol:
            self._ida._ida_bytes.set_cmt(self.address, comment, 0)
        elif comment_type == CommentType.repeatable:
            self._ida._ida_bytes.set_cmt(self.address, comment, 1)
        elif comment_type in (CommentType.anterior, CommentType.plate):
            self._set_extra_comments(self.address, self._ida._ida_lines.E_PREV, comment)
        elif comment_type == CommentType.posterior:
            self._set_extra_comments(self.address, self._ida._ida_lines.E_NEXT, comment)

    @property
    def size(self) -> int:
        return self._ida._ida_bytes.get_item_size(self.address)

    @property
    def type(self) -> LineType:
        flags = self._ida._ida_bytes.get_flags(self.address)

        if self._ida._ida_bytes.is_code(flags):
            return LineType.code

        elif self._ida._ida_bytes.is_byte(flags):
            return LineType.byte

        elif self._ida._ida_bytes.is_word(flags):
            return LineType.word

        elif self._ida._ida_bytes.is_dword(flags):
            return LineType.dword

        elif self._ida._ida_bytes.is_qword(flags):
            return LineType.qword

        elif self._ida._ida_bytes.is_oword(flags):
            return LineType.oword

        elif self._ida._ida_bytes.is_tbyte(flags):
            return LineType.tbyte

        elif self._ida._ida_bytes.is_float(flags):
            return LineType.float

        elif self._ida._ida_bytes.is_double(flags):
            return LineType.double

        elif self._ida._ida_bytes.is_pack_real(flags):
            return LineType.pack_real

        elif self._ida._ida_bytes.is_strlit(flags):
            str_type = self._ida._ida_nalt.get_str_type(self.address)
            str_type &= self._ida._ida_nalt.STRWIDTH_MASK
            if str_type == self._ida._ida_nalt.STRTYPE_C:
                return LineType.string
            elif str_type == self._ida._ida_nalt.STRTYPE_C_16:
                return LineType.string16
            elif str_type == self._ida._ida_nalt.STRTYPE_C_32:
                return LineType.string32
            else:
                RuntimeError(f"String type at {hex(self.address)} not currently supported.")

        elif self._ida._ida_bytes.is_struct(flags):
            return LineType.struct

        elif self._ida._ida_bytes.is_align(flags):
            return LineType.align

        elif self._ida._ida_bytes.is_tail(flags):
            return LineType.tail

        elif self._ida._ida_bytes.is_unknown(flags):
            if self._ida._ida_bytes.is_loaded(self.address):
                return LineType.undefined
            else:
                return LineType.unloaded

        else:
            raise RuntimeError(f"Unexpected line type at {hex(self.address)}")

    @type.setter
    def type(self, new_type: LineType):
        if new_type not in LineType:
            raise ValueError(f"Invalid line type: {new_type}")

        type = self.type
        if type == new_type:
            return  # No change necessary

        # First undefine line, otherwise IDA will fail to change the type in many cases.
        self.undefine()

        # Create new data type at address.
        # NOTE: Using idc over ida_bytes so we don't have to bother with size.
        success = False
        if new_type == LineType.code:
            raise NotImplementedError("Setting a line to code is not currently supported.")

        elif new_type == LineType.byte:
            success = self._ida._idc.create_byte(self.address)

        elif new_type == LineType.word:
            success = self._ida._idc.create_word(self.address)

        elif new_type == LineType.dword:
            success = self._ida._idc.create_dword(self.address)

        elif new_type == LineType.qword:
            success = self._ida._idc.create_qword(self.address)

        elif new_type == LineType.oword:
            success = self._ida._idc.create_oword(self.address)

        elif new_type == LineType.tbyte:
            success = self._ida._idc.create_tbyte(self.address)

        elif new_type == LineType.float:
            success = self._ida._idc.create_float(self.address)

        elif new_type == LineType.double:
            success = self._ida._idc.create_double(self.address)

        elif new_type == LineType.pack_real:
            success = self._ida._idc.create_pack_real(self.address)

        elif new_type == LineType.string:
            success = self._ida._ida_bytes.create_strlit(self.address, 0, self._ida._ida_nalt.STRTYPE_C)

        elif new_type == LineType.string16:
            success = self._ida._ida_bytes.create_strlit(self.address, 0, self._ida._ida_nalt.STRTYPE_C_16)

        elif new_type == LineType.string32:
            success = self._ida._ida_bytes.create_strlit(self.address, 0, self._ida._ida_nalt.STRTYPE_C_32)

        elif new_type == LineType.struct:
            # TODO: Create a special "Structure" class that helps interfacing with structs
            # success = self._ida._idc.create_struct(self.addr, self.size, "TODO")
            raise NotImplementedError("Setting a line to a struct is not currently supported.")

        elif new_type == LineType.align:
            # success = self._ida._idc.create_align(self.addr, 0, 0)
            raise NotImplementedError("Setting a line to an alignment is not currently supported.")

        elif new_type == LineType.tail:
            raise NotImplementedError("Setting a line to a tail is not currently supported.")

        elif new_type == LineType.undefined:
            success = True  # Already undefined, nothing more to do.

        elif new_type == LineType.unloaded:
            # TODO: Should we ever support this?
            raise NotImplementedError("Setting a line to unloaded is not currently supported.")

        if not success:
            raise TypeError(f"Failed to set new type {repr(new_type)} at {hex(self.address)}")

    def undefine(self):
        if self.type in [LineType.undefined, LineType.unloaded]:
            return
        # TODO: kordesii's EncryptedString called with like so. Should we be doing that here?
        #   idc.del_items(self.start_ea, idc.DELIT_SIMPLE, len(self.decoded_data))
        success = self._ida._ida_bytes.del_items(self.address)
        if not success:
            raise TypeError(f"Failed to undefine data at {hex(self.address)}")

    @property
    def value(self) -> Any:
        type = self.type

        if type == LineType.code:
            return IDAInstruction(self._ida, self.address)

        elif type == LineType.byte:
            return self._ida._ida_bytes.get_wide_byte(self.address)

        elif type == LineType.word:
            return self._ida._ida_bytes.get_wide_word(self.address)

        elif type == LineType.dword:
            return self._ida._ida_bytes.get_wide_dword(self.address)

        elif type == LineType.qword:
            return self._ida._ida_bytes.get_qword(self.address)

        elif type == LineType.oword:
            raise NotImplementedError(f"Getting oword value is not currently supported.")

        elif type == LineType.tbyte:
            raise NotImplementedError(f"Getting tbyte value is not currently supported.")

        elif type == LineType.float:
            return self._ida._idc.GetFloat(self.address)

        elif type == LineType.double:
            return self._ida._idc.GetDouble(self.address)

        elif type == LineType.pack_real:
            raise NotImplementedError(f"Getting pack_real value is not currently supported.")

        elif type == LineType.string:
            return self.data.decode("utf-8").rstrip("\x00")

        elif type == LineType.string16:
            return self.data.decode("utf-16-le").rstrip("\x00")

        elif type == LineType.string32:
            return self.data.decode("utf-32-le").rstrip("\x00")

        elif type == LineType.struct:
            raise NotImplementedError(f"Getting struct value is not currently supported.")

        elif type == LineType.align:
            return self.data

        elif type == LineType.tail:
            raise NotImplementedError(f"Getting tail value is not currently supported.")

        elif type == LineType.undefined:
            return self._ida._ida_bytes.get_wide_byte(self.address)

        elif type == LineType.unloaded:
            return None

    @value.setter
    def value(self, new_value: Any):
        matched_types = LineType.match_type(new_value)
        if not matched_types:
            raise TypeError(f"Unsupported value type: {type(new_value)}")

        type_ = self.type

        # First see if setting this new value would required changing the line type.
        if type_ not in matched_types:
            # Use the first entry of matched types to be the type we change it to.
            # Type must be changed after we patch to avoid the type change failing
            # due to bytes originally set at location.
            # (e.g, IDA will fail to change the type to string if original bytes are printable)
            new_type = matched_types[0]
            logger.debug(f"Changing line type from {type_} to {new_type}")
            type_ = new_type

        # TODO: Should we undefine before we patch?

        # Set the value based on line type.
        success = False
        if type_ == LineType.code:
            # TODO: If number of bytes that make up the instruction is lower, make sure to
            #   add nops?
            # self._ida._idc.create_insn(new_value._insn_t)
            raise NotImplementedError(f"Setting instruction value is not currently supported.")

        elif type_ == LineType.byte:
            if isinstance(new_value, bytes):
                new_value = new_value[0]
            success = self._ida._ida_bytes.patch_byte(self.address, new_value)

        elif type_ == LineType.word:
            success = self._ida._ida_bytes.patch_word(self.address, new_value)

        elif type_ == LineType.dword:
            success = self._ida._ida_bytes.patch_dword(self.address, new_value)

        elif type_ == LineType.qword:
            success = self._ida._ida_bytes.patch_qword(self.address, new_value)

        elif type_ == LineType.oword:
            raise NotImplementedError(f"Setting oword value is not currently supported.")

        elif type_ == LineType.tbyte:
            raise NotImplementedError(f"Setting tbyte value is not currently supported.")

        elif type_ == LineType.float:
            new_value_dword = struct.unpack("I", struct.pack("f", new_value))[0]
            success = self._ida._ida_bytes.patch_dword(self.address, new_value_dword)

        elif type_ == LineType.double:
            new_value_qword = struct.unpack("Q", struct.pack("d", new_value))[0]
            success = self._ida._ida_bytes.patch_qword(self.address, new_value_qword)

        elif type_ == LineType.pack_real:
            raise NotImplementedError(f"Setting pack_real value is not currently supported.")

        elif type_ in (LineType.string, LineType.string16, LineType.string32):
            new_value += "\x00"
            encoding = {
                LineType.string: "utf-8",
                LineType.string16: "utf-16-le",
                LineType.string32: "utf-32-le",
            }[type_]
            new_value = new_value.encode(encoding)
            # Ensure there is enough space to fit the new value.
            if len(new_value) > (self.next.address - self.address):
                raise ValueError(f"Encoded string too large to fit on line.")
            # Fill with zero bytes.
            new_value += b"\x00" * (len(self) - len(new_value))
            # For strings, we are going to undefine the value first so ida appropriately
            # resets the item.
            self.undefine()
            self._ida._ida_bytes.patch_bytes(self.address, new_value)
            success = True

        elif type_ == LineType.struct:
            raise NotImplementedError(f"Setting struct value is not currently supported.")

        elif type_ == LineType.align:
            self._ida._ida_bytes.patch_bytes(self.address, new_value)
            success = True

        elif type_ == LineType.tail:
            raise NotImplementedError(f"Getting tail value is not currently supported.")

        elif type_ == LineType.undefined:
            if isinstance(new_value, int):
                new_value = bytes([new_value])
            self._ida._ida_bytes.patch_bytes(self.address, new_value)
            success = True

        elif type_ == LineType.unloaded:
            return  # ignore attempts to set a value to unloaded.

        if success:
            # Now set the new type. (Does nothing if already the right type.)
            self.type = type_
        else:
            raise TypeError(f"Failed to set value {repr(new_value)} at {hex(self.address)}")




from __future__ import annotations

from typing import TYPE_CHECKING, Iterable

from dragodis.ghidra.line import GhidraLine
from dragodis.ghidra.memory import GhidraMemory
from dragodis.interface import Segment, SegmentType, SegmentPermission

if TYPE_CHECKING:
    import ghidra
    from dragodis.ghidra.flat import GhidraFlatAPI


class GhidraSegment(Segment):

    def __init__(self, ghidra: GhidraFlatAPI, memory_block: "ghidra.program.model.mem.MemoryBlock"):
        self._ghidra = ghidra
        self._memory_block = memory_block

    def __contains__(self, addr: int):
        return self._memory_block.contains(self._ghidra._to_addr(addr))

    def __len__(self):
        return self._memory_block.getSize()

    @property
    def name(self) -> str:
        return self._memory_block.getName()

    @property
    def start(self) -> int:
        return self._memory_block.getStart().getOffset()

    @property
    def end(self) -> int:
        return self._memory_block.getEnd().getOffset() + 1

    @property
    def initialized(self) -> bool:
        #TODO can't handle any other address spaces except for the default
        space = self._ghidra._program.getAddressFactory().getDefaultAddressSpace()
        if self._memory_block.getStart().getAddressSpace() != space:
            return False
        return self._memory_block.isInitialized()

    @property
    def bit_size(self) -> int:
        return self._memory_block.getStart().getSize()

    @property
    def permissions(self) -> SegmentPermission:
        permissions = SegmentPermission(0)
        if self._memory_block.isRead():
            permissions |= SegmentPermission.read
        if self._memory_block.isWrite():
            permissions |= SegmentPermission.write
        if self._memory_block.isExecute():
            permissions |= SegmentPermission.execute
        if self._memory_block.isVolatile():
            permissions |= SegmentPermission.volatile
        return permissions

    @property
    def lines(self) -> Iterable[GhidraLine]:
        if self.initialized:
            yield from self._ghidra.lines(self.start, self.end)

    def open(self) -> GhidraMemory:
        if not self.initialized:
            # Empty memory
            return GhidraMemory(self._ghidra, self.start, self.start)
        return GhidraMemory(self._ghidra, self.start, self.end)




from __future__ import annotations
import abc
from functools import lru_cache
from typing import BinaryIO, Optional, Iterable, TYPE_CHECKING

from dragodis.interface.types import SegmentPermission, SegmentType

if TYPE_CHECKING:
    from dragodis.interface import Line, Memory


class Segment(metaclass=abc.ABCMeta):
    """
    Interface for accessing segment information. (Sometimes referred to as 'sections')
    """

    def __str__(self) -> str:
        return f"{self.name}: 0x{self.start:08x} --> 0x{self.end:08x}"

    def __repr__(self):
        return f"<Segment {self} - permissions={self.permissions!r}, bit_size={self.bit_size}>"

    def __contains__(self, addr: int) -> bool:
        """
        Whether the given address is found within the segment.
        Defaults to checking if address is between start and end.
        """
        return self.start <= addr < self.end

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.start == other.start

    def __len__(self) -> int:
        """
        The number of bytes contained in the segment.
        Defaults to checking difference between end and start.
        """
        return self.end - self.start

    @property
    @abc.abstractmethod
    def name(self) -> str:
        """
        The name of the segment.
        """

    @name.setter
    @abc.abstractmethod
    def name(self, new_name: Optional[str]):
        """
        Sets the name of the segment or resets the name if None/"" is provided.
        """

    # TODO: Add ability to set start/end bounds?
    @property
    @abc.abstractmethod
    def start(self) -> int:
        """
        The start address of the segment.
        """

    @property
    @abc.abstractmethod
    def end(self) -> int:
        """
        The end address of the segment (non-inclusive).
        """

    @property
    @abc.abstractmethod
    def initialized(self) -> bool:
        """
        Returns True if we have initialized bytes in the segment.
        """

    @property
    @abc.abstractmethod
    def bit_size(self) -> int:
        """
        The addressing mode in number of bits (8, 16, 32, etc).
        """

    @property
    @abc.abstractmethod
    def permissions(self) -> SegmentPermission:
        """
        The permissions applied to the segment.
        """

    @property
    @abc.abstractmethod
    def lines(self) -> Iterable[Line]:
        """
        Iterates the lines found within the segment.
        """

    @property
    # @lru_cache
    def data(self) -> bytes:
        """
        Raw bytes that make up this segment.
        WARNING: This can consume a lot of memory. Use open() if you would like to
        iteratively stream the contained data instead.

        Defaults to consuming all of the data stream.
        """
        if not self.initialized:
            return b""
        with self.open() as stream:
            return stream.read()

    def get_bytes(self, addr: int, length: int) -> bytes:
        """
        Gets the specified number of bytes at the specified address.

        :param int addr: Address to pull bytes from
        :param int length: Number of bytes to return
        :return: Raw bytes from the segment
        :raises IOError: If address is not within the segment.
        """
        if not self.initialized:
            return b""
        with self.open() as stream:
            stream.seek_address(addr)
            return stream.read(length)

    @abc.abstractmethod
    def open(self) -> Memory:
        """
        Returns an open memory window to access the underlying data in the segment.
        """



from __future__ import annotations
import logging
from typing import List, Optional, TYPE_CHECKING

from dragodis.exceptions import NotExistError
from dragodis.ghidra.operand import GhidraOperand, GhidraARMOperand, Ghidrax86Operand
from dragodis.ghidra.utils import convert_flow_type
from dragodis.interface.instruction import Instruction, x86Instruction, ARMInstruction

if TYPE_CHECKING:
    from dragodis.interface.types import FlowType
    from dragodis.ghidra.flat import GhidraFlatAPI
    import ghidra


logger = logging.getLogger(__name__)


class GhidraInstruction(Instruction):
    _Operand = GhidraOperand

    # TODO: Update as necessary.
    _arm_data_root_mnemonics = [
        "ldr",
        "str",
        "stm",
        "swp",
    ]

    def __init__(self, ghidra: GhidraFlatAPI, instruction: "ghidra.program.model.listing.Instruction"):
        super().__init__(ghidra)
        self._ghidra = ghidra
        self._instruction = instruction

    @property
    def address(self) -> int:
        return self._instruction.getAddress().getOffset()

    @property
    def flow_type(self) -> FlowType:
        return convert_flow_type(self._instruction.getFlowType())

    @property
    def mnemonic(self) -> str:
        # Using lower() to keep consistent with IDA.
        return str(self._instruction.getMnemonicString()).lower()

    @property
    def root_mnemonic(self) -> str:
        # For ARM architectures, pull the second constructor object to get the true
        # base mnemonic without prefixes.
        if self._ghidra.processor_name == "ARM":
            from ghidra.app.plugin.processors.sleigh import SleighDebugLogger
            sleigh_logger = SleighDebugLogger(
                self._ghidra._program,
                self._instruction.getAddress(),
                SleighDebugLogger.SleighDebugMode.MASKS_ONLY
            )
            params = sleigh_logger.getConstructorLineNumbers()
            if len(params) >= 2:
                root_mnem, found, _ = params[1].partition("(")
                if not found:
                    return self.mnemonic
                root_mnem = root_mnem.lower()

                # Some instructions will still include postfixes, such as "ldrb" instead of "ldr".
                # We'll need to manually remove these.
                for root in self._arm_data_root_mnemonics:
                    if root_mnem.startswith(root):
                        return root

                return root_mnem

        return self.mnemonic

    @property
    def operands(self) -> List[GhidraOperand]:
        # HACK: Ignore if operands are implied based on mnemonic.
        # (This is done to better match how IDA does it.)
        if "ES:" in self.text:
            logger.debug(f"Ignoring implied operands at 0x%X: %s", self.address, self.text)
            return []
        return [
            self._Operand(self._ghidra, self, index)
            for index in range(self._instruction.getNumOperands())
        ]

    @property
    def text(self) -> str:
        return str(self._instruction)

    @property
    def stack_depth(self) -> int:
        from ghidra.app.cmd.function import CallDepthChangeInfo
        addr = self._instruction.getAddress()
        func = self._ghidra._listing.getFunctionContaining(addr)
        info = CallDepthChangeInfo(func)
        depth = info.getDepth(addr)
        if depth == func.UNKNOWN_STACK_DEPTH_CHANGE:
            depth = 0
        return depth

    @property
    def stack_delta(self) -> int:
        delta = 0
        flow_type = self._instruction.getFlowType()
        if flow_type.isCall() and flow_type.isUnConditional():
            # the delta we are looking for is actually stored in the function and not the instruction
            flows = self._instruction.getFlows()
            if not flows:
                return delta
            if len(flows) == 1:
                func = self._ghidra._flatapi.getFunctionAt(flows[0])
                # If function has a call fixup, getStackPurgeSize() will be wrong.
                # In this case, we will just calculate from the next instruction's stack depth.
                if func.getCallFixup():
                    next_insn = self.line.next.instruction
                    delta = next_insn.stack_depth - self.stack_depth
                else:
                    delta = func.getStackPurgeSize()
            else:
                logger.warning(f'unexpected number of flows: {flows}')
        else:
            from ghidra.app.cmd.function import CallDepthChangeInfo
            addr = self._instruction.getAddress()
            func = self._ghidra._listing.getFunctionContaining(addr)
            info = CallDepthChangeInfo(func)
            delta = info.getInstructionStackDepthChange(self._instruction)
            if delta == func.UNKNOWN_STACK_DEPTH_CHANGE:
                delta = 0
        return delta


class GhidraARMInstruction(GhidraInstruction, ARMInstruction):
    _Operand = GhidraARMOperand


class Ghidrax86Instruction(GhidraInstruction, x86Instruction):
    _Operand = Ghidrax86Operand

    @property
    def mnemonic(self) -> str:
        mnemonic = super().mnemonic
        # Strip off .rep* prefix if there.
        mnemonic, _, _ = mnemonic.partition(".rep")
        return mnemonic

    @property
    def rep(self) -> Optional[str]:
        if self.data[0] in (0xF2, 0xF3):
            mnemonic = str(self._instruction.getMnemonicString()).lower()
            # Existence of byte prefix doesn't necessarily mean we have a rep prefix.
            # Double confirm by looking at the text itself.
            if ".rep" in mnemonic:
                _, _, rep = mnemonic.partition(".")
                return rep
        return None


GhidraInstruction._ARMInstruction = GhidraARMInstruction
GhidraInstruction._x86Instruction = Ghidrax86Instruction




from __future__ import annotations
from typing import Optional, Iterable, TYPE_CHECKING

from dragodis.ida.reference import IDAReference
from dragodis.interface.symbol import Symbol, Import, Export

if TYPE_CHECKING:
    from dragodis.ida.flat import IDAFlatAPI


class IDASymbol(Symbol):

    def __init__(self, ida: IDAFlatAPI, address: int, name: str):
        self._ida = ida
        self._address = address
        self._name = name

    @property
    def address(self) -> int:
        return self._address

    @property
    def name(self) -> str:
        return self._name

    @property
    def references_to(self) -> Iterable[IDAReference]:
        yield from self._ida.references_to(self.address)


class IDAImport(Import, IDASymbol):

    def __init__(self, ida: IDAFlatAPI, address: int, thunk_address: Optional[int], name: str, namespace: Optional[str]):
        super().__init__(ida, address, name)
        self._thunk_address = thunk_address
        self._namespace = namespace

    @property
    def thunk_address(self) -> Optional[int]:
        return self._thunk_address

    @property
    def namespace(self) -> Optional[str]:
        return self._namespace

    @property
    def references_to(self) -> Iterable[IDAReference]:
        # Pull references to original address pointer.
        thunk_address = self.thunk_address
        for ref in super().references_to:
            # ignore self-references to thunk address.
            if ref.from_address != thunk_address:
                yield ref
        # Also pull references to thunk function if it exists.
        if thunk_address:
            yield from self._ida.references_to(thunk_address)


class IDAExport(Export, IDASymbol):
    ...




from __future__ import annotations
import abc
import os
import pathlib
from typing import Union, Callable

from dragodis.interface.types import ProcessorType


class BackendDisassembler(metaclass=abc.ABCMeta):
    """
    Disassembler Interface

    Handles the setup and teardown of a backend disassembler.
    Should be combined as a mixin with a FlatAPI class.
    """

    # This should be filled in.
    name: str

    # Common processor types dragodis supports.
    PROCESSOR_ARM: str
    PROCESSOR_ARM64: str
    PROCESSOR_X86: str
    PROCESSOR_X64: str

    def __init__(self, input_path: Union[str, pathlib.Path], processor: Union[ProcessorType, str] = None):
        """
        Initialization method.

        This  must perform any initialization, setup, or preparation needed to begin making
        regular calls to the disassembler. For instance, executing the
        disassembler binary, performing initial auto-analysis, and/or
        starting a communications server would all be appropriate actions to
        perform here.

        :param input_path: The path of the file to process
        :param processor: Processor spec to use. (defaults to auto-detection by underlying disassembler)

        :raises NotInstalledError: If the disassembler was not installed on the system.
        """
        self.input_path = pathlib.Path(input_path).resolve()
        if processor:
            processor = self._get_processor_spec(processor)
        self._processor = processor

    def __enter__(self):
        """
        Entry method for context manager interface.
        Used to create a context manager for the disassembler.
        """
        self.start()
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        """
        Exit method for context manager interface.
        Used to create a context manager for the disassembler.
        """
        self.stop(*(exc_type, exc_value, exc_traceback))

    def _get_processor_spec(self, processor: Union[ProcessorType, str]) -> str:
        """
        Generate the appropriate processor spec for the underlying disassembler.
        """
        if isinstance(processor, ProcessorType):
            return {
                ProcessorType.ARM: self.PROCESSOR_ARM,
                ProcessorType.ARM64: self.PROCESSOR_ARM64,
                ProcessorType.x86: self.PROCESSOR_X86,
                ProcessorType.x64: self.PROCESSOR_X64,
            }[processor]
        return processor

    def start(self):
        """
        Setup method.

        This should be called before starting analysis. Any setup should be done here.

        This can include spawning a process to handle the disassembler, creating a temporary
        workspace for the disassembler, etc...
        """
        pass

    def stop(self, *exc_info):
        """
        Teardown method.

        This should be called upon completion of analysis. Any teardown should be done here.

        This can include saving a generated file (such as an IDB), deleting temporary directories,
        shutting down spawned processes, etc...
        """
        pass

    def teleport(self, func: Callable) -> Callable:
        """
        Teleports function into the underlying disassembler to do disassembler specific things.
        NOTE: This function only applies for certain disassemblers. For others, this does nothing
        but pass back the same function.

        e.g.
            def my_func(addr):
                import ida_funcs
                func = ida_funcs.get_func(addr)
                return func.start_ea

            if dis.name == "IDA":
                start_addr = dis.teleport(my_func)(0x123)

        :param func: Callback function to run in underlying disassembler.
        :return: A new function which when called will be run in the underlying disassembler.
        """
        return func



class DragodisError(Exception):
    """Base exception for Dragodis exceptions."""
    pass


class NotInstalledError(DragodisError):
    """
    Raised when trying to use open_program() on a disassembler not installed.
    """


class NotExistError(DragodisError):
    """
    Raised when a disassembler call returns the disassembler specific value
    designating that the desired value does not exist.

    e.g. Trying to get a function with an address that is not part of a function.
    """


class UnsupportedError(DragodisError):
    """
    Raised when a feature is not supported by the current dissassembler.
    """
